.sourceview {
  padding: 0px 12px;
  padding-bottom: 8px;
  padding-left: 0px;
  font-size: 1.4em;

}

.info-view {
  padding: 0px;
  padding-bottom: 4px;
}

.equation-symbol {
  padding: 0px 12px;
}

window > grid {
  padding: 12px;
}


.math-buttons button {
}

  .math-buttons .number-button {
  }

  .math-buttons .numeric-point-button {
  }

  .math-buttons .fx-button {
  }

.bit-toggle-button {
  padding: 1px;
  min-width: 9px;
}

.bit-marker-label {
}

.history-view {
  border-width: 0 0 1px 0;
}

.history-view, .history-entry {
  background-color: shade(@theme_base_color, .95);
}

.history-view:backdrop, .history-entry:backdrop {
  background-color: shade(@theme_unfocused_base_color, .97);
}

.history-entry {
  border-top: 1px groove @borders;
  padding: 6px 3px;
  font-size: 1.3em;
}

  .history-entry .equation-label {
    /* word-spacing: 2px; Doesn't work (yet) in GTK */
  }

  .history-entry .equation-symbol {
    font-weight: lighter;
  }

  .history-entry .answer-label {
  }

/* Used for MathVariablePopover */
row.popover-row {
  border-bottom: 1px groove @borders;
  min-height: 35px;
}
